@@ -1,4 +1,4 @@
-(*$ #use "cinaps_helpers" $*)
+(*$ open Ppxlib_cinaps_helpers $*)
 open! Import
 open Common
 
@@ -69,10 +69,12 @@ module Rule = struct
       | Constant           : Constant.t                                             t
       | Attr_str_type_decl : (structure_item, type_declaration) Attr_group_inline.t t
       | Attr_sig_type_decl : (signature_item, type_declaration) Attr_group_inline.t t
+      | Attr_str_module_type_decl : (structure_item, module_type_declaration) Attr_inline.t t
+      | Attr_sig_module_type_decl : (signature_item, module_type_declaration) Attr_inline.t t
       | Attr_str_type_ext  : (structure_item, type_extension) Attr_inline.t         t
       | Attr_sig_type_ext  : (signature_item, type_extension) Attr_inline.t         t
-      | Attr_str_exception : (structure_item, extension_constructor) Attr_inline.t  t
-      | Attr_sig_exception : (signature_item, extension_constructor) Attr_inline.t  t
+      | Attr_str_exception : (structure_item, type_exception) Attr_inline.t         t
+      | Attr_sig_exception : (signature_item, type_exception) Attr_inline.t         t
 
     type (_, _) equality = Eq : ('a, 'a) equality | Ne : (_, _) equality
 
@@ -87,6 +89,8 @@ module Rule = struct
       | Attr_sig_type_ext  , Attr_sig_type_ext  -> Eq
       | Attr_str_exception , Attr_str_exception -> Eq
       | Attr_sig_exception , Attr_sig_exception -> Eq
+      | Attr_str_module_type_decl, Attr_str_module_type_decl -> Eq
+      | Attr_sig_module_type_decl, Attr_sig_module_type_decl -> Eq
       | _ -> Ne
   end
 
@@ -139,6 +143,14 @@ module Rule = struct
     T (Attr_sig_type_decl, T { attribute; expand; expect = false })
   ;;
 
+  let attr_str_module_type_decl attribute expand =
+    T (Attr_str_module_type_decl, T { attribute; expand; expect = false })
+  ;;
+
+  let attr_sig_module_type_decl attribute expand =
+    T (Attr_sig_module_type_decl, T { attribute; expand; expect = false })
+  ;;
+
   let attr_str_type_ext attribute expand =
     T (Attr_str_type_ext, T { attribute; expand; expect = false })
   ;;
@@ -163,6 +175,14 @@ module Rule = struct
     T (Attr_sig_type_decl, T { attribute; expand; expect = true })
   ;;
 
+  let attr_str_module_type_decl_expect attribute expand =
+    T (Attr_str_module_type_decl, T { attribute; expand; expect = true })
+  ;;
+
+  let attr_sig_module_type_decl_expect attribute expand =
+    T (Attr_sig_module_type_decl, T { attribute; expand; expect = true })
+  ;;
+
   let attr_str_type_ext_expect attribute expand =
     T (Attr_str_type_ext, T { attribute; expand; expect = true })
   ;;
@@ -321,7 +341,10 @@ let handle_attr_group_inline attrs rf items ~loc ~base_ctxt =
       match get_group group.attribute items with
       | None -> acc
       | Some values ->
-        let ctxt = Expansion_context.Deriver.make ~derived_item_loc:loc ~base:base_ctxt () in
+        let ctxt =
+          Expansion_context.Deriver.make ~derived_item_loc:loc
+            ~inline:group.expect ~base:base_ctxt ()
+        in
         let expect_items = group.expand ~ctxt rf items values in
         expect_items :: acc)
 
@@ -331,7 +354,10 @@ let handle_attr_inline attrs item ~loc ~base_ctxt =
       match Attribute.get a.attribute item with
       | None -> acc
       | Some value ->
-        let ctxt = Expansion_context.Deriver.make ~derived_item_loc:loc ~base:base_ctxt () in
+        let ctxt =
+          Expansion_context.Deriver.make ~derived_item_loc:loc
+            ~inline:a.expect ~base:base_ctxt ()
+        in
         let expect_items = a.expand ~ctxt item value in
         expect_items :: acc)
 
@@ -379,6 +405,17 @@ class map_top_down ?(expect_mismatch_handler=Expect_mismatch_handler.nop)
     |> Rule.Attr_group_inline.split_normal_and_expect
   in
 
+  let attr_str_module_type_decls, attr_str_module_type_decls_expect =
+    Rule.filter Attr_str_module_type_decl rules
+    |> sort_attr_inline
+    |> Rule.Attr_inline.split_normal_and_expect
+  in
+  let attr_sig_module_type_decls, attr_sig_module_type_decls_expect =
+    Rule.filter Attr_sig_module_type_decl rules
+    |> sort_attr_inline
+    |> Rule.Attr_inline.split_normal_and_expect
+  in
+
   let attr_str_type_exts, attr_str_type_exts_expect =
     Rule.filter Attr_str_type_ext rules
     |> sort_attr_inline
@@ -461,13 +498,14 @@ class map_top_down ?(expect_mismatch_handler=Expect_mismatch_handler.nop)
        - func.pexp_desc = Pexp_ident _
     *)
     method private pexp_apply_without_traversing_function base_ctxt e func args =
-      let { pexp_desc = _; pexp_loc; pexp_attributes } = e in
+      let { pexp_desc = _; pexp_loc; pexp_attributes; pexp_loc_stack; } = e in
       let func =
-        let { pexp_desc; pexp_loc; pexp_attributes } = func in
+        let { pexp_desc; pexp_loc; pexp_attributes; pexp_loc_stack } = func in
         let pexp_attributes = self#attributes base_ctxt pexp_attributes in
         { pexp_desc
         ; pexp_loc (* location doesn't need to be traversed *)
         ; pexp_attributes
+        ; pexp_loc_stack
         }
       in
       let args = List.map args ~f:(fun (lab, exp) -> (lab, self#expression base_ctxt exp)) in
@@ -475,6 +513,7 @@ class map_top_down ?(expect_mismatch_handler=Expect_mismatch_handler.nop)
       { pexp_loc
       ; pexp_attributes
       ; pexp_desc = Pexp_apply (func, args)
+      ; pexp_loc_stack
       }
 
     method! class_type base_ctxt x =
@@ -527,13 +566,14 @@ class map_top_down ?(expect_mismatch_handler=Expect_mismatch_handler.nop)
         if not in_generated_code then
           Generated_code_hook.insert_after hook Structure_item item.pstr_loc
             (Many extra_items);
+        let original_rest = rest in
         let rest = loop rest ~in_generated_code in
         (match expect_items with
         | [] -> ()
         | _ ->
           let expected = rev_concat expect_items in
           let pos = item.pstr_loc.loc_end in
-          Code_matcher.match_structure rest ~pos ~expected
+          Code_matcher.match_structure original_rest ~pos ~expected
             ~mismatch_handler:(fun loc repl ->
               expect_mismatch_handler.f Structure_item loc repl));
         item :: (extra_items @ rest)
@@ -569,6 +609,15 @@ class map_top_down ?(expect_mismatch_handler=Expect_mismatch_handler.nop)
             in
             with_extra_items item ~extra_items ~expect_items ~rest ~in_generated_code
 
+          | Pstr_modtype mtd ->
+            let extra_items =
+              handle_attr_inline attr_str_module_type_decls mtd ~loc ~base_ctxt
+            in
+            let expect_items =
+              handle_attr_inline attr_str_module_type_decls_expect mtd ~loc ~base_ctxt
+            in
+            with_extra_items item ~extra_items ~expect_items ~rest ~in_generated_code
+
           | Pstr_typext te ->
             let extra_items = handle_attr_inline attr_str_type_exts te ~loc ~base_ctxt in
             let expect_items =
@@ -598,13 +647,14 @@ class map_top_down ?(expect_mismatch_handler=Expect_mismatch_handler.nop)
         if not in_generated_code then
           Generated_code_hook.insert_after hook Signature_item item.psig_loc
             (Many extra_items);
+        let original_rest = rest in
         let rest = loop rest ~in_generated_code in
         (match expect_items with
         | [] -> ()
         | _ ->
           let expected = rev_concat expect_items in
           let pos = item.psig_loc.loc_end in
-          Code_matcher.match_signature rest ~pos ~expected
+          Code_matcher.match_signature original_rest ~pos ~expected
             ~mismatch_handler:(fun loc repl ->
               expect_mismatch_handler.f Signature_item loc repl));
         item :: (extra_items @ rest)
@@ -640,6 +690,15 @@ class map_top_down ?(expect_mismatch_handler=Expect_mismatch_handler.nop)
             in
             with_extra_items item ~extra_items ~expect_items ~rest ~in_generated_code
 
+          | Psig_modtype mtd ->
+            let extra_items =
+              handle_attr_inline attr_sig_module_type_decls mtd ~loc ~base_ctxt
+            in
+            let expect_items =
+              handle_attr_inline attr_sig_module_type_decls_expect mtd ~loc ~base_ctxt
+            in
+            with_extra_items item ~extra_items ~expect_items ~rest ~in_generated_code
+
           | Psig_typext te ->
             let extra_items = handle_attr_inline attr_sig_type_exts te ~loc ~base_ctxt in
             let expect_items =
