@@ -213,10 +213,12 @@ module Deriver = struct
       { name          : string
       ; str_type_decl : (structure, rec_flag * type_declaration list) Generator.t option
       ; str_type_ext  : (structure, type_extension                  ) Generator.t option
-      ; str_exception : (structure, extension_constructor           ) Generator.t option
+      ; str_exception : (structure, type_exception                  ) Generator.t option
+      ; str_module_type_decl : (structure, module_type_declaration  ) Generator.t option
       ; sig_type_decl : (signature, rec_flag * type_declaration list) Generator.t option
       ; sig_type_ext  : (signature, type_extension                  ) Generator.t option
-      ; sig_exception : (signature, extension_constructor           ) Generator.t option
+      ; sig_exception : (signature, type_exception                  ) Generator.t option
+      ; sig_module_type_decl : (signature, module_type_declaration  ) Generator.t option
       ; extension     : (loc:Location.t -> path:string -> core_type -> expression) option
       }
   end
@@ -226,9 +228,11 @@ module Deriver = struct
       { str_type_decl : string list
       ; str_type_ext  : string list
       ; str_exception : string list
+      ; str_module_type_decl : string list
       ; sig_type_decl : string list
       ; sig_type_ext  : string list
       ; sig_exception : string list
+      ; sig_module_type_decl : string list
       }
   end
 
@@ -251,6 +255,10 @@ module Deriver = struct
     let str_exception = { kind = Str; name = "exception"
                         ; get     = (fun t -> t.str_exception)
                         ; get_set = (fun t -> t.str_exception) }
+    let str_module_type_decl =
+      { kind = Str; name = "module type"
+      ; get     = (fun t -> t.str_module_type_decl)
+      ; get_set = (fun t -> t.str_module_type_decl) }
     let sig_type_decl = { kind = Sig; name = "signature type"
                         ; get     = (fun t -> t.sig_type_decl)
                         ; get_set = (fun t -> t.sig_type_decl) }
@@ -260,6 +268,10 @@ module Deriver = struct
     let sig_exception = { kind = Sig; name = "signature exception"
                         ; get     = (fun t -> t.sig_exception)
                         ; get_set = (fun t -> t.sig_exception) }
+    let sig_module_type_decl =
+      { kind = Sig; name = "signature module type"
+      ; get     = (fun t -> t.sig_module_type_decl)
+      ; get_set = (fun t -> t.sig_module_type_decl) }
   end
 
   type t =
@@ -370,9 +382,11 @@ module Deriver = struct
         ?str_type_decl
         ?str_type_ext
         ?str_exception
+        ?str_module_type_decl
         ?sig_type_decl
         ?sig_type_ext
         ?sig_exception
+        ?sig_module_type_decl
         ?extension
         name
     =
@@ -381,9 +395,11 @@ module Deriver = struct
       ; str_type_decl
       ; str_type_ext
       ; str_exception
+      ; str_module_type_decl
       ; sig_type_decl
       ; sig_type_ext
       ; sig_exception
+      ; sig_module_type_decl
       ; extension
       }
     in
@@ -402,9 +418,11 @@ module Deriver = struct
         ?str_type_decl
         ?str_type_ext
         ?str_exception
+        ?str_module_type_decl
         ?sig_type_decl
         ?sig_type_ext
         ?sig_exception
+        ?sig_module_type_decl
         set
     =
     let alias : Alias.t =
@@ -415,9 +433,11 @@ module Deriver = struct
       { str_type_decl = get str_type_decl
       ; str_type_ext  = get str_type_ext
       ; str_exception = get str_exception
+      ; str_module_type_decl = get str_module_type_decl
       ; sig_type_decl = get sig_type_decl
       ; sig_type_ext  = get sig_type_ext
       ; sig_exception = get sig_exception
+      ; sig_module_type_decl = get sig_module_type_decl
       }
     in
     Ppx_derivers.register name (T (Alias alias));
@@ -496,62 +516,74 @@ let mk_deriving_attr context ~prefix ~suffix =
     (fun x -> x)
 ;;
 
-module Attr = struct
-  let suffix = ""
-  let td = mk_deriving_attr ~prefix:"ppxlib." ~suffix Type_declaration
-  let te = mk_deriving_attr ~prefix:"ppxlib." ~suffix Type_extension
-  let ec = mk_deriving_attr ~prefix:"ppxlib." ~suffix Extension_constructor
-
-  module Expect = struct
-    let suffix = "_inline"
-    let td = mk_deriving_attr ~prefix:"ppxlib." ~suffix Type_declaration
-    let te = mk_deriving_attr ~prefix:"ppxlib." ~suffix Type_extension
-    let ec = mk_deriving_attr ~prefix:"ppxlib." ~suffix Extension_constructor
-  end
-end
-
 (* +-----------------------------------------------------------------+
-   | Unused warning stuff                                            |
+   | Unused warning stuff + locations check silencing                |
    +-----------------------------------------------------------------+ *)
 
-let disable_unused_warning_attribute ~loc =
-  ({ txt = "ocaml.warning"; loc },
-   PStr [pstr_eval ~loc (estring ~loc "-32") []])
+let disable_unused_warning_attribute =
+  let loc = Location.none in
+  { attr_name = { txt = "ocaml.warning"; loc };
+    attr_payload = PStr [pstr_eval ~loc (estring ~loc "-32") []];
+    attr_loc = loc; }
 ;;
 
-let inline_doc_attr ~loc =
-  ({ txt = "ocaml.doc"; loc },
-   PStr [pstr_eval ~loc (estring ~loc "@inline") []])
+let inline_doc_attr =
+  let loc = Location.none in
+  { attr_name = { txt = "ocaml.doc"; loc };
+    attr_payload = PStr [pstr_eval ~loc (estring ~loc "@inline") []];
+    attr_loc = loc; }
 ;;
 
-let disable_unused_warning_str ~loc st =
-  if keep_w32_impl () then
+let wrap_str ~loc ~hide st =
+  let include_infos = include_infos ~loc (pmod_structure ~loc st) in
+  let pincl_attributes =
+    if hide then
+      [ inline_doc_attr; Merlin_helpers.hide_attribute ]
+    else
+      [ inline_doc_attr ]
+  in
+  [pstr_include ~loc {include_infos with pincl_attributes}]
+
+let wrap_str ~loc ~hide st =
+  let loc = { loc with loc_ghost = true } in
+  let wrap, st =
+    if keep_w32_impl () then
+      hide, st
+    else if not !do_insert_unused_warning_attribute then
+      hide, Ignore_unused_warning.add_dummy_user_for_values#structure st
+    else
+      (* note: a structure is created because it is not currently possible to
+         attach an [@@ocaml.warning] attribute to a single structure item. *)
+      true, (pstr_attribute ~loc disable_unused_warning_attribute :: st)
+  in
+  if wrap then
+    wrap_str ~loc ~hide st
+  else
     st
-  else if not !do_insert_unused_warning_attribute then
-    Ignore_unused_warning.add_dummy_user_for_values#structure st
-  else
-    (* note: a structure is created because it is not currently possible to
-       attach an [@@ocaml.warning] attribute to a single structure item. *)
-    let include_infos =
-      include_infos ~loc
-        (pmod_structure ~loc
-           (pstr_attribute ~loc (disable_unused_warning_attribute ~loc) :: st))
-    in
-    [pstr_include ~loc
-       {include_infos with pincl_attributes = [inline_doc_attr ~loc]}]
 ;;
 
-let disable_unused_warning_sig ~loc sg =
-  if keep_w32_intf () then
-    sg
+let wrap_sig ~loc ~hide st =
+  let include_infos = include_infos ~loc (pmty_signature ~loc st) in
+  let pincl_attributes =
+    if hide then
+      [ inline_doc_attr; Merlin_helpers.hide_attribute ]
+    else
+      [ inline_doc_attr ]
+  in
+  [psig_include ~loc {include_infos with pincl_attributes}]
+
+let wrap_sig ~loc ~hide sg =
+  let loc = { loc with loc_ghost = true } in
+  let wrap, sg =
+    if keep_w32_intf () then
+      hide, sg
+    else
+      true, (psig_attribute ~loc disable_unused_warning_attribute :: sg)
+  in
+  if wrap then
+    wrap_sig ~loc ~hide sg
   else
-    let include_infos =
-      include_infos ~loc
-        (pmty_signature ~loc
-           (psig_attribute ~loc (disable_unused_warning_attribute ~loc) :: sg))
-    in
-    [psig_include ~loc
-       { include_infos with pincl_attributes = [inline_doc_attr ~loc]}]
+    sg
 ;;
 
 (* +-----------------------------------------------------------------+
@@ -616,73 +648,106 @@ let expand_str_type_decls ~ctxt rec_flag tds values =
     types_used_by_deriving tds
     @ Generator.apply_all ~ctxt (rec_flag, tds) generators;
   in
-  disable_unused_warning_str ~loc:(Expansion_context.Deriver.derived_item_loc ctxt) generated
+  wrap_str ~loc:(Expansion_context.Deriver.derived_item_loc ctxt)
+    ~hide:(not @@ Expansion_context.Deriver.inline ctxt) generated
 
 let expand_sig_type_decls ~ctxt rec_flag tds values =
   let generators = merge_generators Deriver.Field.sig_type_decl values in
   let generated = Generator.apply_all ~ctxt (rec_flag, tds) generators in
-  disable_unused_warning_sig ~loc:(Expansion_context.Deriver.derived_item_loc ctxt) generated
+  wrap_sig ~loc:(Expansion_context.Deriver.derived_item_loc ctxt)
+    ~hide:(not @@ Expansion_context.Deriver.inline ctxt) generated
+
+let expand_str_module_type_decl ~ctxt mtd generators =
+  let generators = Deriver.resolve_all Deriver.Field.str_module_type_decl generators in
+  let generated = Generator.apply_all ~ctxt mtd generators in
+  wrap_str ~loc:(Expansion_context.Deriver.derived_item_loc ctxt)
+    ~hide:(not @@ Expansion_context.Deriver.inline ctxt) generated
+
+let expand_sig_module_type_decl ~ctxt mtd generators =
+  let generators = Deriver.resolve_all Deriver.Field.sig_module_type_decl generators in
+  let generated = Generator.apply_all ~ctxt mtd generators in
+  wrap_sig ~loc:(Expansion_context.Deriver.derived_item_loc ctxt)
+    ~hide:(not @@ Expansion_context.Deriver.inline ctxt) generated
 
 let expand_str_exception ~ctxt ec generators =
   let generators = Deriver.resolve_all Deriver.Field.str_exception generators in
   let generated = Generator.apply_all ~ctxt ec generators in
-  disable_unused_warning_str ~loc:(Expansion_context.Deriver.derived_item_loc ctxt) generated
+  wrap_str ~loc:(Expansion_context.Deriver.derived_item_loc ctxt)
+    ~hide:(not @@ Expansion_context.Deriver.inline ctxt) generated
 
 let expand_sig_exception ~ctxt ec generators =
   let generators = Deriver.resolve_all Deriver.Field.sig_exception generators in
   let generated = Generator.apply_all ~ctxt ec generators in
-  disable_unused_warning_sig ~loc:(Expansion_context.Deriver.derived_item_loc ctxt) generated
+  wrap_sig ~loc:(Expansion_context.Deriver.derived_item_loc ctxt)
+    ~hide:(not @@ Expansion_context.Deriver.inline ctxt) generated
 
 let expand_str_type_ext ~ctxt te generators =
   let generators = Deriver.resolve_all Deriver.Field.str_type_ext generators in
   let generated = Generator.apply_all ~ctxt te generators in
-  disable_unused_warning_str ~loc:(Expansion_context.Deriver.derived_item_loc ctxt) generated
+  wrap_str ~loc:(Expansion_context.Deriver.derived_item_loc ctxt)
+    ~hide:(not @@ Expansion_context.Deriver.inline ctxt) generated
 
 let expand_sig_type_ext ~ctxt te generators =
   let generators = Deriver.resolve_all Deriver.Field.sig_type_ext generators in
   let generated = Generator.apply_all ~ctxt te generators in
-  disable_unused_warning_sig ~loc:(Expansion_context.Deriver.derived_item_loc ctxt) generated
+  wrap_sig ~loc:(Expansion_context.Deriver.derived_item_loc ctxt)
+    ~hide:(not @@ Expansion_context.Deriver.inline ctxt) generated
+
+let rules ~typ ~expand_sig ~expand_str ~rule_str ~rule_sig ~rule_str_expect
+      ~rule_sig_expect =
+  let prefix = "ppxlib." in
+  let deriving_attr = mk_deriving_attr ~suffix:"" ~prefix typ in
+  let deriving_attr_expect = mk_deriving_attr ~suffix:"_inline" ~prefix typ in
+  [ rule_sig deriving_attr expand_sig
+  ; rule_str deriving_attr expand_str
+  ; rule_str_expect deriving_attr_expect expand_str
+  ; rule_sig_expect deriving_attr_expect expand_sig
+  ]
+
+let rules_type_decl =
+  rules ~typ:Type_declaration
+    ~expand_str:expand_str_type_decls
+    ~expand_sig:expand_sig_type_decls
+    ~rule_str:Context_free.Rule.attr_str_type_decl
+    ~rule_sig:Context_free.Rule.attr_sig_type_decl
+    ~rule_str_expect:Context_free.Rule.attr_str_type_decl_expect
+    ~rule_sig_expect:Context_free.Rule.attr_sig_type_decl_expect
+
+let rules_type_ext =
+  rules ~typ:Type_extension
+    ~expand_str:expand_str_type_ext
+    ~expand_sig:expand_sig_type_ext
+    ~rule_str:Context_free.Rule.attr_str_type_ext
+    ~rule_sig:Context_free.Rule.attr_sig_type_ext
+    ~rule_str_expect:Context_free.Rule.attr_str_type_ext_expect
+    ~rule_sig_expect:Context_free.Rule.attr_sig_type_ext_expect
+
+let rules_exception =
+  rules ~typ:Type_exception
+    ~expand_str:expand_str_exception
+    ~expand_sig:expand_sig_exception
+    ~rule_str:Context_free.Rule.attr_str_exception
+    ~rule_sig:Context_free.Rule.attr_sig_exception
+    ~rule_str_expect:Context_free.Rule.attr_str_exception_expect
+    ~rule_sig_expect:Context_free.Rule.attr_sig_exception_expect
+
+let rules_module_type_decl =
+  rules ~typ:Module_type_declaration
+    ~expand_str:expand_str_module_type_decl
+    ~expand_sig:expand_sig_module_type_decl
+    ~rule_str:Context_free.Rule.attr_str_module_type_decl
+    ~rule_sig:Context_free.Rule.attr_sig_module_type_decl
+    ~rule_str_expect:Context_free.Rule.attr_str_module_type_decl_expect
+    ~rule_sig_expect:Context_free.Rule.attr_sig_module_type_decl_expect
 
 let () =
-  Driver.register_transformation "deriving"
-    ~aliases:["type_conv"]
-    ~rules:[ Context_free.Rule.attr_str_type_decl
-               Attr.td
-               expand_str_type_decls
-           ; Context_free.Rule.attr_sig_type_decl
-               Attr.td
-               expand_sig_type_decls
-           ; Context_free.Rule.attr_str_type_ext
-               Attr.te
-               expand_str_type_ext
-           ; Context_free.Rule.attr_sig_type_ext
-               Attr.te
-               expand_sig_type_ext
-           ; Context_free.Rule.attr_str_exception
-               Attr.ec
-               expand_str_exception
-           ; Context_free.Rule.attr_sig_exception
-               Attr.ec
-               expand_sig_exception
-
-           (* [@@deriving_inline] *)
-           ; Context_free.Rule.attr_str_type_decl_expect
-               Attr.Expect.td
-               expand_str_type_decls
-           ; Context_free.Rule.attr_sig_type_decl_expect
-               Attr.Expect.td
-               expand_sig_type_decls
-           ; Context_free.Rule.attr_str_type_ext_expect
-               Attr.Expect.te
-               expand_str_type_ext
-           ; Context_free.Rule.attr_sig_type_ext_expect
-               Attr.Expect.te
-               expand_sig_type_ext
-           ; Context_free.Rule.attr_str_exception_expect
-               Attr.Expect.ec
-               expand_str_exception
-           ; Context_free.Rule.attr_sig_exception_expect
-               Attr.Expect.ec
-               expand_sig_exception
-           ]
+  let rules =
+    [ rules_type_decl
+    ; rules_type_ext
+    ; rules_exception
+    ; rules_module_type_decl
+    ]
+    |> List.concat
+  in
+  Driver.register_transformation "deriving" ~aliases:["type_conv"] ~rules
 ;;
