@@ -1,4 +1,4 @@
-(*$ #use "cinaps_helpers" $*)
+(*$ open Ppxlib_cinaps_helpers $*)
 open Import
 open Utils
 
@@ -10,8 +10,10 @@ let args = ref []
 
 let add_arg key spec ~doc = args := (key, spec, doc) :: !args
 
+let loc_fname = ref None
 let perform_checks = ref Options.perform_checks
 let perform_checks_on_extensions = ref Options.perform_checks_on_extensions
+let perform_locations_check = ref Options.perform_locations_check
 let debug_attribute_drop = ref false
 let apply_list = ref None
 let preprocessor = ref None
@@ -163,8 +165,13 @@ module Transform = struct
           hook.f context loc (Many l));
       (header, footer)
     in
-    let map_impl st =
+    let map_impl st_with_attrs =
       let st =
+        let attrs, st =
+          List.split_while st_with_attrs ~f:(function
+            | { pstr_desc = Pstr_attribute _; _ } -> true
+            | _ -> false)
+        in
         let header, footer =
           match enclose_impl with
           | None   -> ([], [])
@@ -174,17 +181,21 @@ module Transform = struct
         in
         let file_path = File_path.get_default_path_str st in
         let base_ctxt = Expansion_context.Base.top_level ~omp_config ~file_path in
+        let attrs = map#structure base_ctxt attrs in
         let st = map#structure base_ctxt st in
-        match header, footer with
-        | [], [] -> st
-        | _      -> List.concat [ header; st; footer ]
+        List.concat [ attrs; header; st; footer ]
       in
       match impl with
       | None -> st
       | Some f -> f st
     in
-    let map_intf sg =
+    let map_intf sg_with_attrs =
       let sg =
+        let attrs, sg =
+          List.split_while sg_with_attrs ~f:(function
+            | { psig_desc = Psig_attribute _; _ } -> true
+            | _ -> false)
+        in
         let header, footer =
           match enclose_intf with
           | None   -> ([], [])
@@ -194,10 +205,9 @@ module Transform = struct
         in
         let file_path = File_path.get_default_path_sig sg in
         let base_ctxt = Expansion_context.Base.top_level ~omp_config ~file_path in
+        let attrs = map#signature base_ctxt attrs in
         let sg = map#signature base_ctxt sg in
-        match header, footer with
-        | [], [] -> sg
-        | _      -> List.concat [ header; sg; footer ]
+        List.concat [ attrs; header; sg; footer ]
       in
       match intf with
       | None -> sg
@@ -479,7 +489,7 @@ let real_map_structure config cookies st =
     match lint_errors with
     | [] -> st
     | _  ->
-      List.map lint_errors ~f:(fun (({ loc; _ }, _) as attr) ->
+      List.map lint_errors ~f:(fun ({ attr_name = { loc; _ }; _} as attr) ->
         Ast_builder.Default.pstr_attribute ~loc attr)
       @ st
   in
@@ -490,6 +500,11 @@ let real_map_structure config cookies st =
     Attribute.check_unused#structure st;
     if !perform_checks_on_extensions then Extension.check_unused#structure st;
     Attribute.check_all_seen ();
+    if !perform_locations_check then
+      let open Location_check in
+      ignore (
+        (enforce_invariants !loc_fname)#structure
+          st Non_intersecting_ranges.empty : Non_intersecting_ranges.t)
   end;
   st
 ;;
@@ -522,7 +537,7 @@ let real_map_signature config cookies sg =
     match lint_errors with
     | [] -> sg
     | _  ->
-      List.map lint_errors ~f:(fun (({ loc; _ }, _) as attr) ->
+      List.map lint_errors ~f:(fun ({ attr_name = { loc; _ }; _} as attr) ->
         Ast_builder.Default.psig_attribute ~loc attr)
       @ sg
   in
@@ -533,6 +548,11 @@ let real_map_signature config cookies sg =
     Attribute.check_unused#signature sg;
     if !perform_checks_on_extensions then Extension.check_unused#signature sg;
     Attribute.check_all_seen ();
+    if !perform_locations_check then
+      let open Location_check in
+      ignore (
+        (enforce_invariants !loc_fname)#signature
+          sg Non_intersecting_ranges.empty : Non_intersecting_ranges.t)
   end;
   sg
 ;;
@@ -733,7 +753,7 @@ type output_mode =
 (*$*)
 let extract_cookies_str st =
   match st with
-  | { pstr_desc = Pstr_attribute({txt = "ocaml.ppx.context"; _}, _); _ } as prefix
+  | { pstr_desc = Pstr_attribute {attr_name={txt = "ocaml.ppx.context"; _}; _}; _ } as prefix
     :: st ->
     let prefix = Ppxlib_ast.Selected_ast.to_ocaml Structure [prefix] in
     assert (List.is_empty
@@ -751,7 +771,7 @@ let add_cookies_str st =
 (*$ str_to_sig _last_text_block *)
 let extract_cookies_sig sg =
   match sg with
-  | { psig_desc = Psig_attribute({txt = "ocaml.ppx.context"; _}, _); _ } as prefix
+  | { psig_desc = Psig_attribute {attr_name={txt = "ocaml.ppx.context"; _}; _}; _ } as prefix
     :: sg ->
     let prefix = Ppxlib_ast.Selected_ast.to_ocaml Signature [prefix] in
     assert (List.is_empty
@@ -956,7 +976,6 @@ let process_file (kind : Kind.t) fn ~input_name ~relocate ~output_mode ~embed_er
     end
 ;;
 
-let loc_fname = ref None
 let output_mode = ref Pretty_print
 let output = ref None
 let kind = ref None
@@ -1067,6 +1086,10 @@ let shared_args =
     " Disable checks on extension point only"
   ; "-check-on-extensions", Arg.Set perform_checks_on_extensions,
     " Enable checks on extension point only"
+  ; "-no-locations-check", Arg.Clear perform_locations_check,
+    " Disable locations check only"
+  ; "-locations-check", Arg.Set perform_locations_check,
+    " Enable locations check only"
   ; "-apply", Arg.String handle_apply,
     "<names> Apply these transformations in order (comma-separated list)"
   ; "-dont-apply", Arg.String handle_dont_apply,
@@ -1290,5 +1313,13 @@ let () =
        { A.default_mapper with structure; signature })
 
 let enable_checks () =
+  (* We do not enable the locations check here, we currently require that one
+     to be specifically enabled. *)
   perform_checks := true;
   perform_checks_on_extensions := true
+
+let enable_location_check () =
+  perform_locations_check := true
+
+let disable_location_check () =
+  perform_locations_check := false
